@@ -19,10 +19,9 @@ from __future__ import print_function
 
 __author__ = 'A. Jesse Jiryu Davis'
 __email__ = 'jesse@mongodb.com'
-__version__ = '1.3.0'
+__version__ = '1.4.1'
 
 import atexit
-import collections
 import contextlib
 import errno
 import functools
@@ -45,6 +44,11 @@ try:
 except ImportError:
     from Queue import Queue, Empty
 
+try:
+    from collections.abc import Mapping
+except:
+    from collections import Mapping
+
 try:
     from collections import OrderedDict
 except:
@@ -83,13 +87,13 @@ __all__ = [
     'MockupDB', 'go', 'going', 'Future', 'wait_until', 'interactive_server',
 
     'OP_REPLY', 'OP_UPDATE', 'OP_INSERT', 'OP_QUERY', 'OP_GET_MORE',
-    'OP_DELETE', 'OP_KILL_CURSORS',
+    'OP_DELETE', 'OP_KILL_CURSORS', 'OP_MSG',
 
     'QUERY_FLAGS', 'UPDATE_FLAGS', 'INSERT_FLAGS', 'DELETE_FLAGS',
-    'REPLY_FLAGS',
+    'REPLY_FLAGS', 'OP_MSG_FLAGS',
 
     'Request', 'Command', 'OpQuery', 'OpGetMore', 'OpKillCursors', 'OpInsert',
-    'OpUpdate', 'OpDelete', 'OpReply',
+    'OpUpdate', 'OpDelete', 'OpReply', 'OpMsg',
 
     'Matcher', 'absent',
 ]
@@ -239,6 +243,7 @@ OP_QUERY = 2004
 OP_GET_MORE = 2005
 OP_DELETE = 2006
 OP_KILL_CURSORS = 2007
+OP_MSG = 2013
 
 QUERY_FLAGS = OrderedDict([
     ('TailableCursor', 2),
@@ -263,7 +268,13 @@ REPLY_FLAGS = OrderedDict([
     ('CursorNotFound', 1),
     ('QueryFailure', 2)])
 
+OP_MSG_FLAGS = OrderedDict([
+    ('checksumPresent', 1),
+    ('moreToCome', 2)])
+
+_UNPACK_BYTE = struct.Struct("<b").unpack
 _UNPACK_INT = struct.Struct("<i").unpack
+_UNPACK_UINT = struct.Struct("<I").unpack
 _UNPACK_LONG = struct.Struct("<q").unpack
 
 
@@ -273,6 +284,49 @@ def _get_c_string(data, position):
     return _utf_8_decode(data[position:end], None, True)[0], end + 1
 
 
+def _bson_values_equal(a, b):
+    # Check if values are either from our vendored bson or PyMongo's bson.
+    for value in (a, b):
+        if not hasattr(value, '_type_marker'):
+            # Normal equality.
+            return a == b
+
+    def marker(obj):
+        return getattr(obj, '_type_marker', None)
+
+    if marker(a) != marker(b):
+        return a == b
+
+    # Instances of Binary, ObjectId, etc. from our vendored bson don't equal
+    # instances of PyMongo's bson classes that users pass in as message specs,
+    # since isinstance() fails. Reimplement equality checks for each class.
+    key_fn = {
+        # Binary.
+        5: lambda obj: (obj.subtype, bytes(obj)),
+        # ObjectId.
+        7: lambda obj: obj.binary,
+        # Regex.
+        11: lambda obj: (obj.pattern, obj.flags),
+        # Code.
+        13: lambda obj: (obj.scope, str(obj)),
+        # Timestamp.
+        17: lambda obj: (obj.time, obj.inc),
+        # Decimal128.
+        19: lambda obj: obj.bid,
+        # DBRef.
+        100: lambda obj: (obj.database, obj.collection, obj.id),
+        # MaxKey.
+        127: lambda obj: 127,
+        # MinKey.
+        255: lambda obj: 255,
+    }.get(marker(a))
+
+    if key_fn:
+        return key_fn(a) == key_fn(b)
+
+    return a == b
+
+
 class _PeekableQueue(Queue):
     """Only safe from one consumer thread at a time."""
     _NO_ITEM = object()
@@ -298,7 +352,7 @@ class _PeekableQueue(Queue):
 
 
 class Request(object):
-    """Base class for `Command`, `OpInsert`, and so on.
+    """Base class for `Command`, `OpMsg`, and so on.
 
     Some useful asserts you can do in tests:
 
@@ -310,13 +364,13 @@ class Request(object):
     True
     >>> {'_id': 1} == OpInsert([{'_id': 0}, {'_id': 1}])[1]
     True
-    >>> 'field' in Command(field=1)
+    >>> 'field' in OpMsg(field=1)
     True
-    >>> 'field' in Command()
+    >>> 'field' in OpMsg()
     False
-    >>> 'field' in Command('ismaster')
+    >>> 'field' in OpMsg('ismaster')
     False
-    >>> Command(ismaster=False)['ismaster'] is False
+    >>> OpMsg(ismaster=False)['ismaster'] is False
     True
     """
     opcode = None
@@ -333,7 +387,7 @@ class Request(object):
         self._verbose = self._server and self._server.verbose
         self._server_port = kwargs.pop('server_port', None)
         self._docs = make_docs(*args, **kwargs)
-        if not all(isinstance(doc, collections.Mapping) for doc in self._docs):
+        if not all(isinstance(doc, Mapping) for doc in self._docs):
             raise_args_err()
 
     @property
@@ -457,7 +511,7 @@ class Request(object):
                 if value is absent:
                     if key in other_doc:
                         return False
-                elif other_doc.get(key, None) != value:
+                elif not _bson_values_equal(value, other_doc.get(key, None)):
                     return False
             if isinstance(doc, (OrderedDict, _bson.SON)):
                 if not isinstance(other_doc, (OrderedDict, _bson.SON)):
@@ -511,6 +565,130 @@ class Request(object):
         return '%s(%s)' % (name, ', '.join(str(part) for part in parts))
 
 
+class CommandBase(Request):
+    """A command the client executes on the server."""
+    is_command = True
+
+    # Check command name case-insensitively.
+    _non_matched_attrs = Request._non_matched_attrs + ('command_name', )
+
+    @property
+    def command_name(self):
+        """The command name or None.
+
+        >>> OpMsg({'count': 'collection'}).command_name
+        'count'
+        >>> OpMsg('aggregate', 'collection', cursor=absent).command_name
+        'aggregate'
+        """
+        if self.docs and self.docs[0]:
+            return list(self.docs[0])[0]
+
+    def _matches_docs(self, docs, other_docs):
+        assert len(docs) == len(other_docs) == 1
+        doc, = docs
+        other_doc, = other_docs
+        items = list(doc.items())
+        other_items = list(other_doc.items())
+
+        # Compare command name case-insensitively.
+        if items and other_items:
+            if items[0][0].lower() != other_items[0][0].lower():
+                return False
+            if not _bson_values_equal(items[0][1], other_items[0][1]):
+                return False
+        return super(CommandBase, self)._matches_docs(
+            [OrderedDict(items[1:])],
+            [OrderedDict(other_items[1:])])
+
+
+class OpMsg(CommandBase):
+    """An OP_MSG request the client executes on the server."""
+    opcode = OP_MSG
+    is_command = True
+    _flags_map = OP_MSG_FLAGS
+
+    @classmethod
+    def unpack(cls, msg, client, server, request_id):
+        """Parse message and return an `OpMsg`.
+
+        Takes the client message as bytes, the client and server socket objects,
+        and the client request id.
+        """
+        flags, = _UNPACK_UINT(msg[:4])
+        pos = 4
+        first_payload_type, = _UNPACK_BYTE(msg[pos:pos+1])
+        pos += 1
+        first_payload_size, = _UNPACK_INT(msg[pos:pos+4])
+        if flags != 0 and flags != 2:
+            raise ValueError('OP_MSG flag must be 0 or 2 not %r' % (flags,))
+        if first_payload_type != 0:
+            raise ValueError('First OP_MSG payload type must be 0 not %r' % (
+                first_payload_type,))
+
+        # Parse the initial document and add the optional payload type 1.
+        payload_document = _bson.decode_all(msg[pos:pos+first_payload_size],
+                                            CODEC_OPTIONS)[0]
+        pos += first_payload_size
+        if len(msg) != pos:
+            payload_type, = _UNPACK_BYTE(msg[pos:pos+1])
+            pos += 1
+            if payload_type != 1:
+                raise ValueError('Second OP_MSG payload type must be 1 not %r'
+                                 % (payload_type,))
+            section_size, = _UNPACK_INT(msg[pos:pos+4])
+            if len(msg) != pos + section_size:
+                raise ValueError('More than two OP_MSG sections unsupported')
+            pos += 4
+            identifier, pos = _get_c_string(msg, pos)
+            documents = _bson.decode_all(msg[pos:], CODEC_OPTIONS)
+            payload_document[identifier] = documents
+
+        database = payload_document['$db']
+        return OpMsg(payload_document, namespace=database, flags=flags,
+                     _client=client, request_id=request_id,
+                     _server=server)
+
+    def __init__(self, *args, **kwargs):
+        super(OpMsg, self).__init__(*args, **kwargs)
+        if len(self._docs) > 1:
+            raise_args_err('OpMsg too many documents', ValueError)
+
+    @property
+    def slave_ok(self):
+        """True if this OpMsg can read from a secondary."""
+        read_preference = self.doc.get('$readPreference')
+        return read_preference and read_preference.get('mode') != 'primary'
+
+    slave_okay = slave_ok
+    """Synonym for `.slave_ok`."""
+
+    @property
+    def command_name(self):
+        """The command name or None.
+
+        >>> OpMsg({'count': 'collection'}).command_name
+        'count'
+        >>> OpMsg('aggregate', 'collection', cursor=absent).command_name
+        'aggregate'
+        """
+        if self.docs and self.docs[0]:
+            return list(self.docs[0])[0]
+
+    def _replies(self, *args, **kwargs):
+        if self.flags & OP_MSG_FLAGS['moreToCome']:
+            assert False, "Cannot reply to OpMsg with moreToCome: %r" % (self,)
+        reply = make_op_msg_reply(*args, **kwargs)
+        if not reply.docs:
+            reply.docs = [{'ok': 1}]
+        else:
+            if len(reply.docs) > 1:
+                raise ValueError('OP_MSG reply with multiple documents: %s'
+                                 % (reply.docs, ))
+            reply.doc.setdefault('ok', 1)
+        super(OpMsg, self)._replies(reply)
+
+
 class OpQuery(Request):
     """A query (besides a command) the client executes on the server.
 
@@ -555,7 +733,7 @@ class OpQuery(Request):
 
     def __init__(self, *args, **kwargs):
         fields = kwargs.pop('fields', None)
-        if fields is not None and not isinstance(fields, collections.Mapping):
+        if fields is not None and not isinstance(fields, Mapping):
             raise_args_err()
         self._fields = fields
         self._num_to_skip = kwargs.pop('num_to_skip', None)
@@ -592,41 +770,8 @@ class OpQuery(Request):
         return rep + ')'
 
 
-class Command(OpQuery):
+class Command(CommandBase, OpQuery):
     """A command the client executes on the server."""
-    is_command = True
-
-    # Check command name case-insensitively.
-    _non_matched_attrs = OpQuery._non_matched_attrs + ('command_name', )
-
-    @property
-    def command_name(self):
-        """The command name or None.
-
-        >>> Command({'count': 'collection'}).command_name
-        'count'
-        >>> Command('aggregate', 'collection', cursor=absent).command_name
-        'aggregate'
-        """
-        if self.docs and self.docs[0]:
-            return list(self.docs[0])[0]
-
-    def _matches_docs(self, docs, other_docs):
-        assert len(docs) == len(other_docs) == 1
-        doc, = docs
-        other_doc, = other_docs
-        items = list(doc.items())
-        other_items = list(other_doc.items())
-
-        # Compare command name case-insensitively.
-        if items and other_items:
-            if items[0][0].lower() != other_items[0][0].lower():
-                return False
-            if items[0][1] != other_items[0][1]:
-                return False
-        return super(Command, self)._matches_docs(
-            [OrderedDict(items[1:])],
-            [OrderedDict(other_items[1:])])
 
     def _replies(self, *args, **kwargs):
         reply = make_reply(*args, **kwargs)
@@ -780,23 +925,12 @@ class OpDelete(_LegacyWrite):
                    request_id=request_id, _server=server)
 
 
-class OpReply(object):
+class Reply(object):
     """A reply from `MockupDB` to the client."""
     def __init__(self, *args, **kwargs):
         self._flags = kwargs.pop('flags', 0)
-        self._cursor_id = kwargs.pop('cursor_id', 0)
-        self._starting_from = kwargs.pop('starting_from', 0)
         self._docs = make_docs(*args, **kwargs)
 
-    @property
-    def docs(self):
-        """The reply documents, if any."""
-        return self._docs
-
-    @docs.setter
-    def docs(self, docs):
-        self._docs = make_docs(docs)
-
     @property
     def doc(self):
         """Contents of reply.
@@ -807,6 +941,35 @@ class OpReply(object):
         assert len(self._docs) == 1, '%s has more than one document' % self
         return self._docs[0]
 
+    def __str__(self):
+        return docs_repr(*self._docs)
+
+    def __repr__(self):
+        rep = '%s(%s' % (self.__class__.__name__, self)
+        if self._flags:
+            rep += ', flags=' + '|'.join(
+                name for name, value in REPLY_FLAGS.items()
+                if self._flags & value)
+
+        return rep + ')'
+
+
+class OpReply(Reply):
+    """An OP_REPLY reply from `MockupDB` to the client."""
+    def __init__(self, *args, **kwargs):
+        self._cursor_id = kwargs.pop('cursor_id', 0)
+        self._starting_from = kwargs.pop('starting_from', 0)
+        super(OpReply, self).__init__(*args, **kwargs)
+
+    @property
+    def docs(self):
+        """The reply documents, if any."""
+        return self._docs
+
+    @docs.setter
+    def docs(self, docs):
+        self._docs = make_docs(docs)
+
     def update(self, *args, **kwargs):
         """Update the document. Same as ``dict().update()``.
 
@@ -836,17 +999,53 @@ class OpReply(object):
         message += struct.pack("<i", OP_REPLY)
         return message + data
 
-    def __str__(self):
-        return docs_repr(*self._docs)
+
+class OpMsgReply(Reply):
+    """A OP_MSG reply from `MockupDB` to the client."""
+    def __init__(self, *args, **kwargs):
+        super(OpMsgReply, self).__init__(*args, **kwargs)
+        assert len(self._docs) <= 1, 'OpMsgReply can only have one document'
+
+    @property
+    def docs(self):
+        """The reply documents, if any."""
+        return self._docs
+
+    @docs.setter
+    def docs(self, docs):
+        self._docs = make_docs(docs)
+        assert len(self._docs) == 1, 'OpMsgReply must have one document'
+
+    def update(self, *args, **kwargs):
+        """Update the document. Same as ``dict().update()``.
+
+           >>> reply = OpMsgReply({'ismaster': True})
+           >>> reply.update(maxWireVersion=3)
+           >>> reply.doc['maxWireVersion']
+           3
+           >>> reply.update({'maxWriteBatchSize': 10, 'msg': 'isdbgrid'})
+        """
+        self.doc.update(*args, **kwargs)
+
+    def reply_bytes(self, request):
+        """Take a `Request` and return an OP_MSG message as bytes."""
+        flags = struct.pack("<I", self._flags)
+        payload_type = struct.pack("<b", 0)
+        payload_data = _bson.BSON.encode(self.doc)
+        data = b''.join([flags, payload_type, payload_data])
+
+        reply_id = random.randint(0, 1000000)
+        response_to = request.request_id
+
+        header = struct.pack(
+            "<iiii", 16 + len(data), reply_id, response_to, OP_MSG)
+        return header + data
 
     def __repr__(self):
         rep = '%s(%s' % (self.__class__.__name__, self)
-        if self._starting_from:
-            rep += ', starting_from=%d' % self._starting_from
-
         if self._flags:
             rep += ', flags=' + '|'.join(
-                name for name, value in REPLY_FLAGS.items()
+                name for name, value in OP_MSG_FLAGS.items()
                 if self._flags & value)
 
         return rep + ')'
@@ -1023,12 +1222,12 @@ class MockupDB(object):
         self._autoresponders = []
 
         if auto_ismaster is True:
-            self.autoresponds(Command('ismaster'),
+            self.autoresponds(CommandBase('ismaster'),
                               {'ismaster': True,
                                'minWireVersion': min_wire_version,
                                'maxWireVersion': max_wire_version})
         elif auto_ismaster:
-            self.autoresponds(Command('ismaster'), auto_ismaster)
+            self.autoresponds(CommandBase('ismaster'), auto_ismaster)
 
     @_synchronized
     def run(self):
@@ -1109,9 +1308,9 @@ class MockupDB(object):
         >>> future = go(client.db.command, 'foo')
         >>> s.got('foo')
         True
-        >>> s.got(Command('foo', namespace='db'))
+        >>> s.got(OpMsg('foo', namespace='db'))
         True
-        >>> s.got(Command('foo', key='value'))
+        >>> s.got(OpMsg('foo', key='value'))
         False
         >>> s.ok()
         >>> future() == {'ok': 1}
@@ -1175,17 +1374,20 @@ class MockupDB(object):
 
         The remaining arguments are a :ref:`message spec <message spec>`:
 
+        >>> # ok
         >>> responder = s.autoresponds('bar', ok=0, errmsg='err')
         >>> client.db.command('bar')
         Traceback (most recent call last):
         ...
         OperationFailure: command SON([('bar', 1)]) on namespace db.$cmd failed: err
-        >>> responder = s.autoresponds(Command('find', 'collection'),
+        >>> responder = s.autoresponds(OpMsg('find', 'collection'),
         ...                            {'cursor': {'id': 0, 'firstBatch': [{'_id': 1}, {'_id': 2}]}})
+        >>> # ok
         >>> list(client.db.collection.find()) == [{'_id': 1}, {'_id': 2}]
         True
-        >>> responder = s.autoresponds(Command('find', 'collection'),
+        >>> responder = s.autoresponds(OpMsg('find', 'collection'),
         ...                            {'cursor': {'id': 0, 'firstBatch': [{'a': 1}, {'a': 2}]}})
+        >>> # bad
         >>> list(client.db.collection.find()) == [{'a': 1}, {'a': 2}]
         True
 
@@ -1197,6 +1399,7 @@ class MockupDB(object):
         and replied to. Future matching requests skip the queue.
 
         >>> future = go(client.db.command, 'baz')
+        >>> # bad
         >>> responder = s.autoresponds('baz', {'key': 'value'})
         >>> future() == {'ok': 1, 'key': 'value'}
         True
@@ -1236,7 +1439,7 @@ class MockupDB(object):
         ...         print('logging: %r' % request)
         >>> responder = s.autoresponds(logger)
         >>> client.db.command('baz') == {'ok': 1, 'a': 2}
-        logging: Command({"baz": 1}, flags=SlaveOkay, namespace="db")
+        logging: OpMsg({"baz": 1, "$db": "db", "$readPreference": {"mode": "primaryPreferred"}}, namespace="db")
         True
 
         The synonym `subscribe` better expresses your intent if your handler
@@ -1466,7 +1669,8 @@ def bind_socket(address):
     raise socket.error('could not bind socket')
 
 
-OPCODES = {OP_QUERY: OpQuery,
+OPCODES = {OP_MSG: OpMsg,
+           OP_QUERY: OpQuery,
            OP_INSERT: OpInsert,
            OP_UPDATE: OpUpdate,
            OP_DELETE: OpDelete,
@@ -1511,7 +1715,7 @@ def mock_server_receive(sock, length):
 
 
 def make_docs(*args, **kwargs):
-    """Make the documents for a `Request` or `OpReply`.
+    """Make the documents for a `Request` or `Reply`.
 
     Takes a variety of argument styles, returns a list of dicts.
 
@@ -1537,7 +1741,7 @@ def make_docs(*args, **kwargs):
 
     if isinstance(args[0], (list, tuple)):
         # Send a batch: OpReply([{'a': 1}, {'a': 2}]).
-        if not all(isinstance(doc, (OpReply, collections.Mapping))
+        if not all(isinstance(doc, (OpReply, Mapping))
                    for doc in args[0]):
             raise_args_err('each doc must be a dict:')
         if kwargs:
@@ -1561,7 +1765,7 @@ def make_docs(*args, **kwargs):
         raise_args_err(err_msg, ValueError)
 
     # Send a batch as varargs: OpReply({'a': 1}, {'a': 2}).
-    if not all(isinstance(doc, (OpReply, collections.Mapping)) for doc in args):
+    if not all(isinstance(doc, (OpReply, Mapping)) for doc in args):
         raise_args_err('each doc must be a dict')
 
     return args
@@ -1603,7 +1807,7 @@ def make_prototype_request(*args, **kwargs):
 
 def make_reply(*args, **kwargs):
     # Error we might raise.
-    if args and isinstance(args[0], OpReply):
+    if args and isinstance(args[0], (OpReply, OpMsgReply)):
         if args[1:] or kwargs:
             raise_args_err("can't interpret args")
         return args[0]
@@ -1611,6 +1815,16 @@ def make_reply(*args, **kwargs):
     return OpReply(*args, **kwargs)
 
 
+def make_op_msg_reply(*args, **kwargs):
+    # Error we might raise.
+    if args and isinstance(args[0], (OpReply, OpMsgReply)):
+        if args[1:] or kwargs:
+            raise_args_err("can't interpret args")
+        return args[0]
+
+    return OpMsgReply(*args, **kwargs)
+
+
 def unprefixed(bson_str):
     rep = unicode(repr(bson_str))
     if rep.startswith(u'u"') or rep.startswith(u"u'"):
@@ -1728,7 +1942,7 @@ def interactive_server(port=27017, verbose=True, all_ok=False, name='MockupDB',
     server.autoresponds('whatsmyuri', you='localhost:12345')
     server.autoresponds({'getLog': 'startupWarnings'},
                         log=['hello from %s!' % name])
-    server.autoresponds(Command('buildInfo'), version='MockupDB ' + __version__)
-    server.autoresponds(Command('listCollections'))
+    server.autoresponds(OpMsg('buildInfo'), version='MockupDB ' + __version__)
+    server.autoresponds(OpMsg('listCollections'))
     server.autoresponds('replSetGetStatus', ok=0)
     return server
